import torch
import torch.nn as nn

from torch.distributions import Normal

from .bayesian_model import BayesianModel

class Hyperparameters:
    def __init__(self, dims=[1, 2, 1], bias=2*[True], activations=2*[torch.sigmoid]):
        self.dims = dims
        self.bias = bias
        self.activations = activations

        if len(self.dims) < 3:
            raise ValueError

        if (len(self.dims) != len(self.activations)+1):
            raise ValueError

class MLP(BayesianModel):
    def __init__(self, loss, constraint=None, bounds=[-float('inf'), float('inf')], temperature=None, prior=None,
    hparams=Hyperparameters(), savefile=None, dtype=torch.float64, device='cpu'):
        super().__init__(
            loss=loss, constraint=constraint, bounds=bounds, temperature=temperature, dtype=dtype, device=device)
        self.hp = hparams
        self.fc_layers = self.set_fc_layers()
        self.prior = prior or self.default_prior()
        if savefile:
            self.load_state_dict(savefile, strict=False)

    def default_prior(self):
        return Normal(
            torch.zeros(self.num_params(), dtype=self.dtype, device=self.device),
            torch.ones(self.num_params(), dtype=self.dtype, device=self.device)
        )

    def set_fc_layers(self):
        fc = []
        for i in range(len(self.hp.dims)-1):
            fc.append(nn.Linear(
                self.hp.dims[i], self.hp.dims[i+1], bias=self.hp.bias[i]
            ).to(dtype=self.dtype, device=self.device))
        return nn.ModuleList(fc)

    def forward(self, x):
        for fc, activation in zip(self.fc_layers, self.hp.activations):
            x = fc(x)
            if activation is not None:
                x = activation(x)
        return x

    def num_hidden_layers(self):
        """ Get the number of hidden layers. """
        return len(self.hp.dims)-2



import torch

def squared_mmd(x1, x2, kernel, biased=True):
    n1 = len(x1)
    n2 = len(x2)

    if biased:
        return (kernel.sum_symm_K(x1, include_diag=True) / (n1 ** 2)
            + kernel.sum_symm_K(x2, include_diag=True) / (n2 ** 2)
            - 2 * kernel.sum_K(x1, x2) / (n1 * n2)
        )
    else:
        return (kernel.sum_symm_K(x1, include_diag=False) / (n1 * (n1 - 1))
            + kernel.sum_symm_K(x2, include_diag=False) / (n2 * (n2 - 1))
            - 2 * kernel.sum_K(x1, x2) / (n1 * n2)
        )

def mmd(x1, x2, kernel):
    return torch.sqrt(squared_mmd(x1, x2, kernel, biased=True))



from datetime import timedelta
from pathlib import Path
from timeit import default_timer as timer

from .sampler import Sampler

class SerialSampler(Sampler):
    """ Serial MCMC Sampler """

    def __init__(self, counter):
        self.counter = counter

    def set_verbose_run_msg(self):
        return 'Iteration {:' \
            + str(len(str(self.counter.num_iters))) \
            + '} out of ' \
            + str(self.counter.num_iters) \
            + ' (in epoch {:' \
            + str(len(str(self.counter.num_epochs))) \
            + '} out of ' \
            + str(self.counter.num_epochs) \
            + '), duration {}'

    def set_verbose_benchmark_msg(self, num_chains):
        return 'Simulating chain {:' \
            + str(len(str(num_chains))) \
            + '} out of ' \
            + str(num_chains) \
            + ' ({:' \
            + str(len(str(num_chains))) \
            + '} failures due to not meeting conditions and {:' \
            + str(len(str(num_chains))) \
            + '} failures due to runtime error)...'

    def run(self, num_epochs, num_burnin_epochs, verbose=False, verbose_step=100):
        """ Run the sampler """

        self.counter.set_epoch_info(num_epochs, num_burnin_epochs)
        verbose_msg = self.set_verbose_run_msg()

        for i in range(self.counter.num_epochs):
            for _, (x, y) in enumerate(self.dataloader):
                if verbose and ((self.counter.idx % verbose_step) == 0):
                    start_time = timer()

                self.draw(x, y, savestate=False if (self.counter.idx < self.counter.num_burnin_iters) else True)

                if verbose and (((self.counter.idx+1) % verbose_step) == 0):
                    end_time = timer()
                    print(verbose_msg.format(self.counter.idx+1, i+1, timedelta(seconds=end_time-start_time)))

                self.counter.increment_idx()

    def benchmark(
        self,
        num_chains,
        num_epochs,
        num_burnin_epochs,
        path,
        check_conditions=None,
        verbose=False,
        verbose_step=100,
        print_acceptance=False,
        print_runtime=True
    ):
        if verbose:
            verbose_msg = self.set_verbose_benchmark_msg(num_chains)

        i, j, k = 0, 0, 0

        while i < num_chains:
            if verbose:
                print(verbose_msg.format(i+1, j, k))

            run_path = Path(path).joinpath('run'+str(i+1).zfill(len(str(num_chains))))

            try:
                theta0 = self.get_model().prior.sample()
                self.reset(theta0.clone().detach(), data=None, reset_counter=True, reset_chain=True)

                start_time = timer()
                self.run(
                    num_epochs=num_epochs, num_burnin_epochs=num_burnin_epochs, verbose=verbose, verbose_step=verbose_step
                )
                end_time = timer()
                runtime = end_time - start_time

                if ((check_conditions is None) or check_conditions(self.get_chain(), runtime)):
                    self.get_chain().to_chainfile(path=run_path, mode='w')

                    with open(run_path.joinpath('runtime.txt'), 'w') as file:
                        file.write("{}\n".format(runtime))

                    i = i + 1

                    if verbose:
                        print('Succeeded', end='')
                else:
                    j = j + 1

                    if verbose:
                        print('Failed due to not meeting conditions', end='')

                if verbose:
                    if print_acceptance:
                        print('; acceptance rate = {}'.format(self.get_chain().acceptance_rate()), end='')
                    if print_runtime:
                        print('; runtime = {}'.format(timedelta(seconds=runtime)), end='')
                    print('\n')
            except RuntimeError as error:
                with open(run_path.joinpath('errors', 'error'+str(k+1).zfill(num_chains)), 'w') as file:
                    file.write("{}\n".format(error))

                k = k + 1

                if verbose:
                    print('Failed due to runtime error\n')

        with open(Path(path).joinpath('run_counts.txt'), 'w') as file:
            file.write("{},succesful\n".format(i))
            file.write("{},unmet_conditions\n".format(j))
            file.write("{},runtime_errors\n".format(k))



def binary_cross_entropy(x, y, reduction='mean'):
    loss = -(x.log() * y + (1 - x).log() * (1 - y))

    if reduction == 'mean':
        result = loss.mean()
    elif reduction == 'sum':
        result = loss.sum()
    else:
        raise ValueError

    return result



import torch

from .single_chain_serial_sampler import SingleChainSerialSampler
from eeyore.chains import ChainList
from eeyore.datasets import DataCounter

class RAM(SingleChainSerialSampler):
    def __init__(self, model,
        theta0=None, dataloader=None, data0=None, counter=None,
        cov0=None, a=0.234, g=0.7, chain=ChainList()):
        super(RAM, self).__init__(counter or DataCounter.from_dataloader(dataloader))
        self.model = model
        self.dataloader = dataloader
        self.a = a
        self.g = g

        if cov0 is not None:
            self.cov0 = cov0.clone().detach()
        else:
            self.cov0 = torch.eye(self.model.num_params(), dtype=self.model.dtype, device=self.model.device)
        self.keys = ['sample', 'target_val', 'accepted']
        self.chain = chain

        if theta0 is not None:
            self.set_all(theta0.clone().detach(), data=data0)

    def set_current(self, theta, data=None):
        x, y = super().set_current(theta, data=data)
        self.current['target_val'] = self.model.log_target(self.current['sample'].clone().detach(), x, y)

    def set_cov(self, cov=None):
        self.chol_cov = torch.cholesky(cov or self.cov0)

    def set_all(self, theta, data=None, cov=None):
        super().set_all(theta, data=data)
        self.set_cov(cov=cov)

    def draw(self, x, y, savestate=False, offset=0):
        proposed = {key : None for key in self.keys}

        randn_sample = torch.randn(self.model.num_params(), dtype=self.model.dtype, device=self.model.device)
        proposed['sample'] = self.current['sample'].clone().detach() + self.chol_cov @ randn_sample
        proposed['target_val'] = self.model.log_target(proposed['sample'].clone().detach(), x, y)

        log_rate = proposed['target_val'] - self.current['target_val']

        if torch.log(torch.rand(1, dtype=self.model.dtype, device=self.model.device)) < log_rate:
            self.current['sample'] = proposed['sample'].clone().detach()
            self.current['target_val'] = proposed['target_val'].clone().detach()
            self.current['accepted'] = 1
        else:
            self.model.set_params(self.current['sample'].clone().detach())
            self.current['accepted'] = 0

        h = min(1, self.model.num_params() * (self.counter.idx + 1 - offset) ** (-self.g))
        self.chol_cov = torch.cholesky(self.chol_cov @ (
            torch.eye(self.model.num_params(), dtype=self.model.dtype, device=self.model.device) + \
            h * (min(1, torch.exp(log_rate).item()) - self.a
            ) * torch.ger(randn_sample, randn_sample) / \
            torch.dot(randn_sample, randn_sample).item()) @ self.chol_cov.t())

        if savestate:
            self.chain.detach_and_update(self.current)

        self.current['sample'].detach_()
        self.current['target_val'].detach_()



# See algorithms 4 and 5 in
# https://arxiv.org/abs/1111.4246

import numpy as np

from .tuner import Tuner

class HMCDATuner(Tuner):
    def __init__(self, l, e0=None, d=0.65, eub=None):
        self.l = l
        self.e0 = e0
        self.d = d
        self.eub = eub # Upper bound for leapfrog step in tuning phase, not in original algorithm

        if self.e0 is None:
            self.m = None
        else:
            self.set_m(self.e0)

        if self.eub is None:
            self.logeub = None
        else:
            self.logeub = np.log(self.eub)

        self.logbare = 0.
        self.barh = 0.
        self.g = 0.05
        self.t0 = 10
        self.k = 0.75

    def set_m(self, e0):
        self.m = np.log(10 * e0)

    def set_d_w(self, iter):
        self.d_w = 1 / (iter + self.t0)

    def set_e_w(self, iter):
        self.e_w = 1 / (iter ** self.k)

    def num_steps(self, e):
        return max(1, round(self.l / e))

    def tune(self, rate, idx, return_e=True):
        iter = idx + 1
        self.set_d_w(iter)
        self.set_e_w(iter)

        self.barh = (1 - self.d_w) * self.barh + self.d_w * (self.d - rate)
        loge = self.m - np.sqrt(iter) * self.barh / self.g
        if self.logeub is not None:
            loge = min(loge, self.logeub)
        self.logbare = self.e_w * loge + (1 - self.e_w) * self.logbare

        if return_e:
            e = np.exp(loge)
            return e, self.num_steps(e)
        else:
            bare = np.exp(self.logbare)
            return bare, self.num_steps(bare)



# See lemma 2 of section 4.1 in https://www.tandfonline.com/doi/abs/10.1080/10618600.1998.10474787

import torch

from eeyore.linalg import is_pos_def, nearest_pd

from .cov import cov
from .mc_cov import mc_cov

# x is a numpy array of 3 dimensions, (chain, MC iteration, parameter)
def multi_rhat(x, mc_cov_mat=None, method='inse', adjust=False):
    num_chains, num_iters, num_pars = x.shape

    w = torch.zeros([num_pars, num_pars])
    for i in range(num_chains):
        if mc_cov_mat is None:
            w = w + mc_cov(x[i], method=method, adjust=adjust, rowvar=False)
        else:
            w = w + mc_cov_mat[i]
    w = w / num_chains

    if not is_pos_def(w):
        w = nearest_pd(w)
        is_w_pd = False
    else:
        is_w_pd = True

    b = cov(x.mean(1), rowvar=False)

    if not is_pos_def(b):
        b = nearest_pd(b)
        is_b_pd = False
    else:
        is_b_pd = True

    rhat = torch.eig(torch.matmul(torch.inverse(w), b), eigenvectors=False)[0][:, 0].max().item()
    rhat = ((num_iters - 1) / num_iters) + ((num_chains + 1) / num_chains) * rhat

    return rhat, w, b, is_w_pd, is_b_pd



import torch

class Chain:
    """ Base class for Monte Carlo chains """

    def reset(self):
        raise NotImplementedError

    def update(self, state):
        raise NotImplementedError

    def detach_and_update(self, state):
        self.update({k: v.clone().detach() if isinstance(v, torch.Tensor) else v for k, v in state.items()})



from .data_counter import DataCounter
from .data_info import data_paths
from .empty_dataset import EmptyXYDataset
from .idataset import IDataset
from .xydataset import XYDataset
from .xyidataset import XYIDataset



import numpy as np
import torch

from .single_chain_serial_sampler import SingleChainSerialSampler
from eeyore.chains import ChainList
from eeyore.datasets import DataCounter

class SMMALA(SingleChainSerialSampler):
    def __init__(self, model, theta0=None, dataloader=None, data0=None, counter=None, step=0.1, transform=None,
    chain=ChainList()):
        super(SMMALA, self).__init__(counter or DataCounter.from_dataloader(dataloader))
        self.model = model
        self.dataloader = dataloader
        self.step = step
        self.transform = transform

        self.keys = ['sample', 'target_val', 'grad_val', 'metric_val', 'inv_metric_val', 'first_term_val', 'accepted']
        self.chain = chain

        if theta0 is not None:
            self.set_current(theta0.clone().detach(), data=data0)

    def set_current(self, theta, data=None):
        x, y = super().set_current(theta, data=data)
        self.current['sample'].requires_grad_(True)
        self.current['target_val'], self.current['grad_val'], self.current['metric_val'] = \
            self.model.upto_metric_log_target(self.current['sample'].clone().detach(), x, y)
        if self.transform is not None:
            self.current['metric_val'] = self.transform(self.current['metric_val'])
        self.current['inv_metric_val'] = torch.inverse(self.current['metric_val'])
        self.current['chol_inv_metric_val'] = torch.cholesky(self.current['inv_metric_val'])

        # See 'Riemann manifold Langevin and Hamiltonian Monte Carlo methods'
        # First summand appearing in equation (10) of page 130
        # Product of metric tensor with gradient
        self.current['first_term_val'] = self.current['inv_metric_val'] @ self.current['grad_val']

    def draw(self, x, y, savestate=False):
        proposed = {key : None for key in self.keys}

        proposal_mean = self.current['sample'] + 0.5 * self.step * self.current['first_term_val']

        proposed['sample'] = \
            proposal_mean + np.sqrt(self.step) * self.current['chol_inv_metric_val'] @ \
            torch.randn(self.model.num_params(), dtype=self.model.dtype, device=self.model.device)
        proposed['target_val'], proposed['grad_val'], proposed['metric_val'] = \
            self.model.upto_metric_log_target(proposed['sample'].clone().detach(), x, y)
        if self.transform is not None:
            proposed['metric_val'] = self.transform(proposed['metric_val'])
        proposed['inv_metric_val'] = torch.inverse(proposed['metric_val'])
        proposed['first_term_val'] = proposed['inv_metric_val'] @ proposed['grad_val']

        log_rate = proposed['target_val'] - self.current['target_val']
        loc_minus_proposal_mean = proposed['sample'] - proposal_mean
        inv_metric_sign, inv_metric_logdet = torch.slogdet(self.step * self.current['inv_metric_val'])
        log_rate = \
            log_rate + 0.5 * (inv_metric_logdet + (loc_minus_proposal_mean.t() @ (self.current['metric_val'] @ \
            loc_minus_proposal_mean))/self.step)

        proposal_mean = proposed['sample'] + 0.5 * self.step * proposed['first_term_val']

        loc_minus_proposal_mean = self.current['sample'] - proposal_mean
        inv_metric_sign, inv_metric_logdet = torch.slogdet(self.step * proposed['inv_metric_val'])
        log_rate = \
            log_rate - 0.5 * (inv_metric_logdet +(loc_minus_proposal_mean.t() @ (proposed['metric_val'] @ \
            loc_minus_proposal_mean))/self.step)

        if torch.log(torch.rand(1, dtype=self.model.dtype, device=self.model.device)) < log_rate:
            self.current['sample'] = proposed['sample'].clone().detach()
            self.current['target_val'] = proposed['target_val'].clone().detach()
            self.current['grad_val'] = proposed['grad_val'].clone().detach()
            self.current['metric_val'] = proposed['metric_val'].clone().detach()
            self.current['inv_metric_val'] = proposed['inv_metric_val'].clone().detach()
            self.current['chol_inv_metric_val'] = torch.cholesky(self.current['inv_metric_val']).clone().detach()
            self.current['first_term_val'] = proposed['first_term_val'].clone().detach()
            self.current['accepted'] = 1
        else:
            self.model.set_params(self.current['sample'].clone().detach())
            self.current['accepted'] = 0

        if savestate:
            self.chain.detach_and_update(self.current)

        self.current['sample'].detach_()
        self.current['target_val'].detach_()
        self.current['grad_val'].detach_()
        self.current['metric_val'].detach_()
        self.current['inv_metric_val'].detach_()
        self.current['chol_inv_metric_val'].detach_()
        self.current['first_term_val'].detach_()



import torch

import eeyore.stats as st

from .chain_file import ChainFile

class ChainLists:
    def __init__(self, keys=['sample', 'target_val', 'accepted'], vals=None):
        self.reset(keys=keys, vals=vals)

    def reset(self, keys=['sample', 'target_val', 'accepted'], vals=None):
        if vals is None:
            self.vals = {key : [] for key in keys}
        else:
            self.vals = vals

    @classmethod
    def from_chain_list(selfclass, chain_lists, keys=['sample', 'target_val', 'accepted']):
        common_keys = set.intersection(*[set(chain_list.vals.keys()) for chain_list in chain_lists])
        class_keys = set(keys) & common_keys

        vals = {}

        for key in class_keys:
            vals[key] = [chain_list.vals[key] for chain_list in chain_lists]

        return selfclass(keys=class_keys, vals=vals)

    @classmethod
    def from_file(selfclass, paths, keys=['sample', 'target_val', 'accepted'], mode='a', dtype=torch.float64, device='cpu'):
        chain_lists = []

        for path in paths:
            chain_lists.append(ChainFile(keys=keys, path=path, mode=mode).to_chainlist(dtype=dtype, device=device))

        return selfclass.from_chain_list(chain_lists, keys=keys)

    def __repr__(self):
        return f"{len(self)} Markov chains, each containing {self.num_samples()} samples."

    def __len__(self):
        return self.num_chains()

    def num_params(self):
        return len(self.vals['sample'][0][0])

    def num_samples(self):
        return len(self.vals['sample'][0])

    def num_chains(self):
        return len(self.vals['sample'])

    def get_samples(self):
        return torch.stack([self.get_chain(i, key='sample') for i in range(self.num_chains())])

    def get_target_vals(self):
        return torch.stack([self.get_chain(i, key='target_val') for i in range(self.num_chains())])

    def get_grad_vals(self):
        return torch.stack([self.get_chain(i, key='grad_val') for i in range(self.num_chains())])

    def get_chain(self, idx, key='sample'):
        return torch.stack(self.vals[key][idx])

    def mean(self):
        return self.get_samples().mean(1)

    def mean_summary(self, g=lambda x: torch.mean(x, dim=0)):
        return g(self.mean())

    def mc_se(self, mc_cov_mat=None, method='inse', adjust=False):
        return torch.stack([
            st.mc_se(self.get_chain(i, key='sample'), method=method, adjust=adjust, rowvar=False)
            if mc_cov_mat is None
            else st.mc_se_from_cov(mc_cov_mat[i])
            for i in range(self.num_chains())
        ])

    def mc_se_summary(self, g=lambda x: torch.mean(x, dim=0), mc_cov_mat=None, method='inse', adjust=False):
        return g(self.mc_se(mc_cov_mat=mc_cov_mat, method=method, adjust=adjust))

    def mc_cov(self, method='inse', adjust=False):
        return torch.stack([
            st.mc_cov(self.get_chain(i, key='sample'), method=method, adjust=adjust, rowvar=False)
            for i in range(self.num_chains())
        ])

    def mc_cov_summary(self, g=lambda m: torch.mean(m, dim=0), method='inse', adjust=False):
        return g(self.mc_cov(method=method, adjust=adjust))

    def mc_cor(self, mc_cov_mat=None, method='inse', adjust=False):
        return torch.stack([
            st.mc_cor(self.get_chain(i, key='sample'), method=method, adjust=adjust, rowvar=False)
            if mc_cov_mat is None
            else st.cor_from_cov(mc_cov_mat[i])
            for i in range(self.num_chains())
        ])

    def mc_cor_summary(self, g=lambda m: torch.mean(m, dim=0), mc_cov_mat=None, method='inse', adjust=False):
        return g(self.mc_cor(mc_cov_mat=mc_cov_mat, method=method, adjust=adjust))

    def acceptance(self):
        return [sum(self.vals['accepted'][i]) / self.num_samples() for i in range(self.num_chains())]

    def acceptance_summary(self, g=lambda x: sum(x) / len(x)):
        return g(self.acceptance())

    def multi_ess(self, mc_cov_mat=None, method='inse', adjust=False):
        return [
            st.multi_ess(
                self.get_chain(i, key='sample'),
                mc_cov_mat=None if mc_cov_mat is None else mc_cov_mat[i],
                method=method,
                adjust=adjust
            )
            for i in range(self.num_chains())
        ]

    def multi_ess_summary(self, g=lambda x: sum(x) / len(x), mc_cov_mat=None, method='inse', adjust=False):
        return g(self.multi_ess(mc_cov_mat=mc_cov_mat, method=method, adjust=adjust))

    def multi_rhat(self, mc_cov_mat=None, method='inse', adjust=False):
        return st.multi_rhat(self.get_samples(), mc_cov_mat=mc_cov_mat, method=method, adjust=adjust)

    def summary(
        self,
        keys=['multi_ess', 'multi_rhat'],
        g_mean_summary=lambda x: torch.mean(x, dim=0),
        g_mc_se_summary=lambda x: torch.mean(x, dim=0),
        g_acceptance_summary=lambda x: sum(x) / len(x),
        g_multi_ess_summary=lambda x: sum(x) / len(x),
        mc_cov_mat=None,
        method='inse',
        adjust=False):
        summaries = {}

        if any(item in keys for item in ['mc_se', 'multi_ess', 'multi_rhat']):
            if mc_cov_mat is None:
                mc_cov_mat = self.mc_cov(method=method, adjust=adjust)

        for key in keys:
            if key == 'mean':
                summaries[key] = self.mean_summary(g=g_mean_summary)
            elif key == 'mc_se':
                summaries[key] = self.mc_se_summary(g=g_mc_se_summary, mc_cov_mat=mc_cov_mat, method=method, adjust=adjust)
            elif key == 'acceptance':
                summaries[key] = self.acceptance_summary(g=g_acceptance_summary)
            elif key == 'multi_ess':
                summaries[key] = self.multi_ess_summary(
                    g=g_multi_ess_summary, mc_cov_mat=mc_cov_mat, method=method, adjust=adjust
                )
            elif key == 'multi_rhat':
                summaries[key], _, _, _, _ = self.multi_rhat(mc_cov_mat=mc_cov_mat, method=method, adjust=adjust)

        return summaries



import numpy as np
import torch

from scipy.stats import truncnorm

from .single_chain_serial_sampler import SingleChainSerialSampler
from eeyore.chains import ChainList
from eeyore.datasets import DataCounter
from eeyore.kernels import NormalKernel, TruncatedNormalKernel

class MALA(SingleChainSerialSampler):
    def __init__(self, model,
        theta0=None, dataloader=None, data0=None, counter=None,
        step=0.1, kernel=None, chain=ChainList()):
        super().__init__(counter or DataCounter.from_dataloader(dataloader))
        self.model = model
        self.dataloader = dataloader
        self.step = step

        self.keys = ['sample', 'target_val', 'grad_val', 'accepted']
        self.chain = chain

        if theta0 is not None:
            self.set_current(theta0.clone().detach(), data=data0)

        self.kernel = kernel or self.default_kernel(self.current)

    def set_current(self, theta, data=None):
        x, y = super().set_current(theta, data=data)
        self.current['target_val'], self.current['grad_val'] = \
            self.model.upto_grad_log_target(self.current['sample'].clone().detach(), x, y)

    def kernel_mean(self, state):
        return state['sample'] + 0.5 * self.step * state['grad_val']

    def default_kernel(self, state):
        loc = self.kernel_mean(state)
        scale = torch.full([self.model.num_params()], np.sqrt(self.step), dtype=self.model.dtype, device=self.model.device)
        if (self.model.constraint is None) or (self.model.constraint == 'transformation'):
            return NormalKernel(loc, scale)
        elif self.model.constraint == 'truncation':
            return TruncatedNormalKernel(loc, scale, lower_bound=self.model.bounds[0], upper_bound=self.model.bounds[1])

    def set_kernel(self, state):
        self.kernel.set_density_params(self.kernel_mean(state))

    def draw(self, x, y, savestate=False):
        proposed = {key : None for key in self.keys}

        proposed['sample'] = self.kernel.sample()

        proposed['target_val'], proposed['grad_val'] = \
            self.model.upto_grad_log_target(proposed['sample'].clone().detach(), x, y)

        log_rate = proposed['target_val'] - self.current['target_val']

        log_rate = log_rate - self.kernel.log_prob(proposed['sample'])

        self.set_kernel(proposed)

        log_rate = log_rate + self.kernel.log_prob(self.current['sample'])

        if torch.log(torch.rand(1, dtype=self.model.dtype, device=self.model.device)) < log_rate:
            self.current['sample'] = proposed['sample'].clone().detach()
            self.current['target_val'] = proposed['target_val'].clone().detach()
            self.current['grad_val'] = proposed['grad_val'].clone().detach()
            self.current['accepted'] = 1
        else:
            self.model.set_params(self.current['sample'].clone().detach())
            self.set_kernel(self.current)
            self.current['accepted'] = 0

        if savestate:
            self.chain.detach_and_update(self.current)

        self.current['sample'].detach_()
        self.current['target_val'].detach_()
        self.current['grad_val'].detach_()



import torch

from .homogeneous_kernel import HomogeneousKernel

class IsoSEKernel(HomogeneousKernel):
    """ Isotropic squared exponential kernel """

    def __init__(self, scale=1., l=1.):
        self.scale = scale # square of amplitude
        self.l = l # square of lengthscale

    def k(self, x1, x2):
        return  torch.exp(-self.squared_dist(x1, x2).div(2. * self.l)).mul(self.scale)



# https://github.com/pytorch/pytorch/pull/32377/commits/d4dae5c2d1fbb0f23c69a40bbad2e6dfdf2fa6b7

from torch.distributions.distribution import Distribution

class TruncatedDistribution(Distribution):
    def __init__(self, base_dist, lower_bound=-float('inf'), upper_bound=float('inf'), *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.base_dist = base_dist
        self.z = self.base_dist.cdf(self.upper_bound) - self.base_dist.cdf(self.lower_bound)

    def log_prob(self, x):
        log_prob_val = self.base_dist.log_prob(x) - self.z.log()
        log_prob_val[(x <= self.lower_bound) | (x >= self.upper_bound)] = -float('inf')
        return log_prob_val

    def cdf(self, x):
        return (self.base_dist.cdf(x) - self.base_dist.cdf(self.lower_bound)) / self.z



import hashlib
import torch
import torch.nn as nn

class Model(nn.Module):
    """ Class representing sampleable neural network model """
    def __init__(self, dtype=torch.float64, device='cpu'):
        super().__init__()
        self.dtype = dtype
        self.device = device

    def summary(self, hashsummary=False):
        print(self)
        print("-" * 80)
        n_params = self.num_params()
        print(f"Number of model parameters: {n_params}")
        print("-" * 80)

        if hashsummary:
            print('Hash Summary:')
            for idx, hashvalue in enumerate(self.hashsummary()):
                print(f"{idx}: {hashvalue}")

    def hashsummary(self):
        """ Print a model summary - checksums of each layer parameters """
        children = list(self.children())

        result = []
        for child in children:
            result.extend(hashlib.sha256(x.detach().cpu().numpy().tobytes()).hexdigest() for x in child.parameters())

        return result

    def num_params(self):
        """ Get the number of model parameters. """
        return sum(p.numel() for p in self.parameters())

    def get_params(self):
        return torch.cat([p.view(-1) for p in self.parameters()])

    def get_grad(self):
        return torch.cat([p.grad.view(-1) for p in self.parameters()])

    def set_params(self, theta, grad_val=None):
        """ Set model parameters with theta. """
        i = 0
        for p in self.parameters():
            j = p.numel()
            p.data = theta[i:i+j].view(p.size())
            if p.grad is not None:
                p.grad.detach_()
                p.grad.zero_()
            if grad_val is not None:
                p.grad = grad_val[i:i+j].view(p.size())
            i += j



from torch.distributions import Normal

from .normalized_kernel import NormalizedKernel

class NormalKernel(NormalizedKernel):
    """ Normal kernel """

    def __init__(self, loc, scale):
        self.set_density(loc, scale)

    def set_density(self, loc, scale):
        """ Set normal probability density function """
        self.density = Normal(loc, scale)

    def set_density_params(self, loc, scale=None):
        """ Set the parameters of normal probability density function """
        self.density.loc = loc
        if scale is not None:
            self.density.scale = scale

    def k(self, x1, x2, scale=None):
        self.set_density_params(x2, scale=scale)
        return self.log_prob(x1).exp()



import torch

from torch.distributions import Normal

from .normalized_kernel import NormalizedKernel

class DEMCKernel(NormalizedKernel):
    """ Normal transition kernel for DEMC"""

    def __init__(self, a=None, b=None, c=0.1, density=None):
        self.a = a
        self.b = b
        self.c = c
        self.density = density

    def init_a_and_b(self, n, dtype, device):
        self.a = torch.empty(n, dtype=dtype, device=device)
        self.b = torch.empty(n, dtype=dtype, device=device)

    def init_density(self, n, dtype, device):
        self.density = Normal(torch.empty(n, dtype=dtype, device=device), torch.empty(n, dtype=dtype, device=device))

    def set_a_and_b(self, a, b):
        self.a = a
        self.b = b

    def mean(self, theta):
        return theta + self.c * (self.a - self.b)

    def set_density(self, theta, scale):
        """ Set normal probability density function """
        self.density = Normal(self.mean(theta), scale)

    def set_density_params(self, theta, scale=None):
        """ Set the parameters of of normal probability density function """
        self.density.loc = self.mean(theta)
        if scale is not None:
            self.density.scale = scale



from .cor import cor
from .cor_from_cov import cor_from_cov
from .cov import cov
from .discrepancy import mmd
from .inse_mc_cov import inse_mc_cov
from .loss import binary_cross_entropy
from .mc_cor import mc_cor
from .mc_cov import mc_cov
from .mc_se import mc_se
from .mc_se_from_cov import mc_se_from_cov
from .metrics import softabs
from .multi_ess import multi_ess
from .multi_rhat import multi_rhat
from .random import choose, choose_from_subset
from .recursive_cov import recursive_cov
from .recursive_mean import recursive_mean
from .running_mean import running_mean



import numpy as np
import torch

from pathlib import Path
from torch.nn.functional import one_hot
from torch.utils.data import Dataset

from .data_info import data_paths
from eeyore.constants import torch_to_np_types

class XYDataset(Dataset):
    def __init__(self, x, y):
        self.set_data(x, y)

    def __repr__(self):
        return f'XYDataset'

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]

    def set_data(self, x, y):
        self.x = x
        self.y = y

    @classmethod
    def from_file(selfclass, path=Path.cwd(), xfile='x.csv', yfile='y.csv', xskiprows=1, yskiprows=1, xusecols=None,
        yusecols=None, xndmin=2, yndmin=2, dtype=torch.float64, device='cpu', xonehot=False, yonehot=False):
        x = torch.from_numpy(np.loadtxt(
            path.joinpath(xfile),
            dtype=torch_to_np_types[dtype], delimiter=',', skiprows=xskiprows, usecols=xusecols, ndmin=xndmin
        )).to(device=device)
        if xonehot:
            x = one_hot(x.long()).to(x.dtype)

        y = torch.from_numpy(np.loadtxt(
            path.joinpath(yfile),
            dtype=torch_to_np_types[dtype], delimiter=',', skiprows=yskiprows, usecols=yusecols, ndmin=yndmin
        )).to(device=device)
        if yonehot:
            y = one_hot(y.long()).to(y.dtype)

        return selfclass(x, y)

    @classmethod
    def from_eeyore(selfclass, data_name,
        xndmin=2, yndmin=2, dtype=torch.float64, device='cpu', xonehot=False, yonehot=False):
        return selfclass.from_file(
            path=data_paths[data_name],
            xndmin=xndmin, yndmin=yndmin, dtype=dtype, device=device, xonehot=xonehot, yonehot=yonehot
        )



import torch
import torch.nn as nn

from .log_target_model import LogTargetModel

class DistributionModel(LogTargetModel):
    def __init__(self, log_pdf, num_params, temperature=None, dtype=torch.float64, device='cpu', requires_grad=True):
        super().__init__(constraint=None, temperature=temperature, dtype=dtype, device=device)
        self.log_pdf = log_pdf
        self.theta = nn.Parameter(
            data=torch.empty(num_params, dtype=self.dtype, device=self.device), requires_grad=requires_grad
        )

    def summary(self, hashsummary=False):
        print(self)
        print("-" * 80)
        print(f"Number of distribution parameters: {self.num_params()}")
        print("-" * 80)

    def log_target(self, theta, x, y):
        self.set_params(theta)

        log_target_val = self.log_pdf(self.theta, x, y)

        if self.temperature is not None:
            log_target_val = self.temperature * log_target_val

        return log_target_val



import torch

# https://discuss.pytorch.org/t/covariance-and-gradient-support/16217

def cov(x, rowvar=False):
    if x.dim() > 2:
        raise ValueError('x has more than 2 dimensions')
    if x.dim() < 2:
        x = x.view(1, -1)
    if not rowvar and x.size(0) != 1:
        x = x.t()

    x_ctr = x - torch.mean(x, dim=1, keepdim=True)

    return x_ctr.matmul(x_ctr.t()).squeeze() / (x.size(1) - 1)



from eeyore.distributions import TruncatedNormal

from .normalized_kernel import NormalizedKernel

class TruncatedNormalKernel(NormalizedKernel):
    """ Truncated normal kernel """

    def __init__(self, loc, scale, lower_bound=-float('inf'), upper_bound=float('inf')):
        self.set_density(loc, scale, lower_bound, upper_bound)

    def set_density(self, loc, scale, lower_bound=-float('inf'), upper_bound=float('inf')):
        """ Set truncated normal probability density function """
        self.density = TruncatedNormal(loc, scale, lower_bound=lower_bound, upper_bound=upper_bound)

    def set_density_params(self, loc, scale=None, lower_bound=-float('inf'), upper_bound=float('inf')):
        """ Set the parameters of truncated normal probability density function """
        self.density.base_dist.loc = loc
        if scale is not None:
            self.density.base_dist.scale = scale
        self.density.set_a_b()



# https://link.springer.com/article/10.1007/BF00143942

import torch

from scipy.stats import truncnorm

from torch.distributions import Exponential, Normal, Uniform

from .truncated_distribution import TruncatedDistribution

class TruncatedNormal(TruncatedDistribution):
    def __init__(self, loc, scale, lower_bound=-float('inf'), upper_bound=float('inf'), *args, **kwargs):
        super().__init__(Normal(loc, scale), lower_bound=lower_bound, upper_bound=upper_bound, *args, **kwargs)

        self.set_a_b()

    def set_a(self):
        self.a = (self.lower_bound - self.base_dist.loc) / self.base_dist.scale

    def set_b(self):
        self.b = (self.upper_bound - self.base_dist.loc) / self.base_dist.scale

    def set_a_b(self):
        self.set_a()
        self.set_b()

    def num_params(self):
        return len(self.base_dist.loc)

    def log_prob(self, x):
        return torch.sum(torch.from_numpy(truncnorm.logpdf(
            x.detach().cpu().numpy(),
            a=self.a.detach().cpu().numpy(),
            b=self.b.detach().cpu().numpy(),
            loc=self.base_dist.loc.detach().cpu().numpy(),
            scale=self.base_dist.scale.detach().cpu().numpy()
            )).to(dtype=self.base_dist.loc.dtype, device=self.base_dist.loc.device))

    def sample_lower_bounded(self):
        rate = 0.5 * (self.a + (self.a ** 2 + 4).sqrt())

        sample = []

        for i in range(self.num_params()):
            while True:
                proposed = Exponential(rate[i] / self.base_dist.scale[i]).sample() + self.lower_bound
                ratio = (-0.5 * ((proposed - self.base_dist.loc[i]) / self.base_dist.scale[i] - rate[i]) ** 2).exp()
                if torch.rand(1, dtype=self.base_dist.loc.dtype, device=self.base_dist.loc.device) <= ratio:
                    break

            sample.append(proposed)

        return torch.stack(sample)

    def sample_upper_bounded(self):
        rate = 0.5 * (-self.b + (self.b ** 2 + 4).sqrt())

        sample = []

        for i in range(self.num_params()):
            while True:
                proposed = Exponential(rate[i] / self.base_dist.scale[i]).sample() - self.upper_bound
                ratio = (-0.5 * ((proposed + self.base_dist.loc[i]) / self.base_dist.scale[i] - rate[i]) ** 2).exp()
                if torch.rand(1, dtype=self.base_dist.loc.dtype, device=self.base_dist.loc.device) <= ratio:
                    break

            sample.append(-proposed)

        return torch.stack(sample)

    def sample_doubly_bounded(self):
        sample = []

        for i in range(self.num_params()):
            while True:
                proposed = Uniform(self.a[i], self.b[i]).sample()

                if ((self.a[i] < 0) and (0 < self.b[i])):
                    ratio = (-0.5 * (proposed ** 2)).exp()
                elif self.b[i] < 0:
                    ratio = (0.5 * (self.b[i] ** 2 - proposed ** 2)).exp()
                elif 0 < self.a[i]:
                    ratio = (0.5 * (self.a[i] ** 2 - proposed ** 2)).exp()

                if torch.rand(1, dtype=self.base_dist.loc.dtype, device=self.base_dist.loc.device) <= ratio:
                    break

            sample.append(proposed)

        return self.base_dist.loc + self.base_dist.scale * torch.stack(sample)

    def sample(self):
        if ((self.lower_bound != -float('inf')) and (self.upper_bound != float('inf'))):
            return self.sample_doubly_bounded()
        elif ((self.lower_bound != -float('inf')) and (self.upper_bound == float('inf'))):
            return self.sample_lower_bounded()
        elif ((self.lower_bound == -float('inf')) and (self.upper_bound != float('inf'))):
            return self.sample_upper_bounded()



import torch

from .cov import cov
from .mc_cov import mc_cov

def multi_ess(x, mc_cov_mat=None, method='inse', adjust=False):
    num_iters, num_pars = x.shape

    cov_mat_det = torch.det(cov(x, rowvar=False)).item()
    mc_cov_mat_det = torch.det(
        mc_cov(x, method=method, adjust=adjust, rowvar=False) if mc_cov_mat is None else mc_cov_mat
    ).item()

    return num_iters * ((cov_mat_det / mc_cov_mat_det) ** (1/num_pars))



import numpy as np
import torch

import eeyore.stats as st

from pathlib import Path

from kanga.chains import ChainArray

from .chain import Chain

class ChainList(Chain):
    """ Monte Carlo chain to store samples in lists """

    def __init__(self, keys=['sample', 'target_val', 'accepted'], vals=None):
        self.reset(keys=keys, vals=vals)

    def reset(self, keys=['sample', 'target_val', 'accepted'], vals=None):
        if vals is None:
            self.vals = {key : [] for key in keys}
        else:
            self.vals = vals

    def __repr__(self):
        return f"Markov chain containing {len(self)} samples."

    def __len__(self):
        return self.num_samples()

    def num_params(self):
        return len(self.get_sample(0))

    def num_samples(self):
        return len(self.vals['sample'])

    def get_param(self, idx):
        return torch.stack([sample[idx] for sample in self.vals['sample']])

    def get_sample(self, idx):
        return self.vals['sample'][idx]

    def get_samples(self):
        return torch.stack(self.vals['sample'])

    def get_target_vals(self):
        return torch.stack(self.vals['target_val'])

    def get_grad_val(self, idx):
        return self.vals['grad_val'][idx]

    def get_grad_vals(self):
        return torch.stack(self.vals['grad_val'])

    def state(self, idx=-1):
        current = {}
        for key, val in self.vals.items():
            try:
                current[key] = val[idx]
            except IndexError:
                print(f'WARNING: chain does not have values for {key}.')
                pass
        return current

    def update(self, state):
        """ Update the chain """
        for key in self.vals.keys():
            self.vals[key].append(state[key])

    def mean(self):
        """ Get the mean of the chain's samples """
        return self.get_samples().mean(0)

    def running_mean(self, idx):
        return st.running_mean(self.get_param(idx))

    def running_means(self):
        return st.running_mean(self.get_samples(), dim=0)

    def mc_se(self, mc_cov_mat=None, method='inse', adjust=False):
        if mc_cov_mat is None:
            return st.mc_se(self.get_samples(), method=method, adjust=adjust, rowvar=False)
        else:
            return st.mc_se_from_cov(mc_cov_mat)

    def mc_cov(self, method='inse', adjust=False):
        return st.mc_cov(self.get_samples(), method=method, adjust=adjust, rowvar=False)

    def mc_cor(self, mc_cov_mat=None, method='inse', adjust=False):
        if mc_cov_mat is None:
            return st.mc_cor(self.get_samples(), method=method, adjust=adjust, rowvar=False)
        else:
            return st.cor_from_cov(mc_cov_mat)

    def acceptance_rate(self):
        """ proportion of accepted samples """
        return sum(self.vals['accepted']) / self.num_samples()

    def multi_ess(self, mc_cov_mat=None, method='inse', adjust=False):
        return st.multi_ess(self.get_samples(), mc_cov_mat=mc_cov_mat, method=method, adjust=adjust)

    def save(self, path):
        """ Save the chain to disk """
        torch.save(self.vals, path)

    def load(self, path):
        """ Load a previously saved chain """
        self.vals = torch.load(path)

    def to_chainfile(self, keys=None, path=Path.cwd(), mode='a'):
        from .chain_file import ChainFile

        chainfile = ChainFile(keys=keys or self.vals.keys(), path=path, mode=mode)

        for i in range(len(self)):
            chainfile.update(self.state(i), reset=False, close=False)

        chainfile.close()

    def to_kanga(self, keys=None):
        keys = set(keys or self.vals.keys()) & set(['sample', 'target_val', 'grad_val', 'accepted'])

        vals = {}

        for key in keys:
            if key == 'sample':
                vals[key] = self.get_samples().detach().cpu().numpy()
            elif key == 'target_val':
                vals[key] = self.get_target_vals().detach().cpu().numpy()
            elif key == 'grad_val':
                vals[key] = self.get_grad_vals().detach().cpu().numpy()
            elif key == 'accepted':
                vals[key] = np.array(self.vals['accepted'])

        return ChainArray(vals)



from pathlib import Path

from .serial_sampler import SerialSampler

class MultiChainSerialSampler(SerialSampler):
    """ Serial MCMC Sampler with multiple chains"""
    def __init__(self, counter):
        super().__init__(counter=counter)

    def default_indicator(self):
        return 0

    def get_model(self, idx=None):
        return self.samplers[idx or self.default_indicator()].model

    def get_chain(self, idx=None):
        return self.samplers[idx or self.default_indicator()].chain

    def get_param(self, param_idx, chain_idx=None):
        return self.get_chain(idx=chain_idx).get_param(idx=param_idx)

    def get_sample(self, sample_idx, chain_idx=None):
        return self.get_chain(idx=chain_idx).get_sample(idx=sample_idx)

    def set_current(self, theta, data=None):
        x, y = data or next(iter(self.dataloader))
        for sampler in self.samplers:
            sampler.set_current(theta, data=(x, y))

    def set_all(self, theta, data=None):
        x, y = data or next(iter(self.dataloader))
        for sampler in self.samplers:
            self.set_all(theta, data=(x, y))

    def reset_chains(self):
        for sampler in self.samplers:
            sampler.chain.reset(keys=sampler.chain.vals.keys())

    def reset(self, theta, data=None, reset_counter=True, reset_chain=True):
        x, y = data or next(iter(self.dataloader))
        for sampler in self.samplers:
            sampler.reset(theta, data=(x, y), reset_counter=reset_counter, reset_chain=reset_chain)

    def to_chainfile(self, path=Path.cwd(), mode='a'):
        for i, sampler in enumerate(self.samplers):
            sampler.chain.to_chainfile(path=path.joinpath('sampler'+str(i).zfill(self.num_chains)), mode=mode)



import numpy as np
import torch

from pathlib import Path

from .chain import Chain
from eeyore.constants import torch_to_np_types

class ChainFile(Chain):
    """ Monte Carlo chain to store samples in file """

    def __init__(self, keys=['sample', 'target_val', 'accepted'], path=Path.cwd(), mode='a'):
        self.path = path
        self.mode = mode

        if not self.path.exists():
            self.path.mkdir(parents=True, exist_ok=True)

        self.reset(keys=keys)

    def reset(self, keys=['sample', 'target_val', 'accepted']):
        self.vals = {key : open(self.path.joinpath(key+'.csv'), self.mode) for key in keys}

    def close(self):
        for key in self.vals.keys():
            self.vals[key].close()

    def update(self, state, reset=True, close=True):
        """ Update the chain """
        if reset:
            self.reset(keys=self.vals.keys())

        for key in self.vals.keys():
            if isinstance(state[key], torch.Tensor):
                np.savetxt(self.vals[key], state[key].detach().cpu().numpy().ravel()[np.newaxis], delimiter=',')
            elif isinstance(state[key], np.ndarray):
                np.savetxt(self.vals[key], state[key].ravel()[np.newaxis], delimiter=',')
            else:
                self.vals[key].write(str(state[key])+'\n')

        if close:
            self.close()

    def scalar_line_to_val_element(self, line, dtype=int):
        return dtype(line.strip())

    def singleton_line_to_val_element(self, line, dtype=torch.float64, device='cpu'):
        return torch.tensor(torch_to_np_types[dtype](line.strip())).to(device=device)

    def vector_line_to_val_element(self, line, dtype=torch.float64, device='cpu'):
        return torch.tensor(list(map(torch_to_np_types[dtype], line.split(',')))).to(device=device)

    def line_to_val_element(self, line, key, dtype=torch.float64, device='cpu'):
        if key == 'target_val':
            return self.singleton_line_to_val_element(line, dtype=dtype, device=device)
        elif (key == 'sample') or (key == 'grad_val'):
            return self.vector_line_to_val_element(line, dtype=dtype, device=device)
        elif key == 'accepted':
            return self.scalar_line_to_val_element(line, dtype=int)

    def to_chainlist(self, keys=None, dtype=torch.float64, device='cpu'):
        from .chain_list import ChainList

        keys = set(keys or self.vals.keys()) & set(['sample', 'target_val', 'grad_val', 'accepted'])

        chainlist_keys = []
        chainlist_vals = []

        for key in keys:
            chainlist_keys.append(key)
            with open(self.path.joinpath(key+'.csv'), mode='r') as file:
                chainlist_vals.append([
                    self.line_to_val_element(line, key, dtype=dtype, device=device) for line in file.readlines()
                ])

        chainlist = ChainList(vals=dict(zip(chainlist_keys, chainlist_vals)))

        return chainlist



import torch

from .xydataset import XYDataset

class EmptyXYDataset(XYDataset):
    def __init__(self, dtype=torch.float64, device='cpu'):
        super().__init__(torch.tensor([[]], dtype=dtype, device=device), torch.tensor([[]], dtype=dtype, device=device))

    def __repr__(self):
        return f'Empty XYDataset'



import copy
import torch

from torch.distributions import Bernoulli

from .am import AM
from .mala import MALA
from .metropolis_hastings import MetropolisHastings
from .ram import RAM
from .single_chain_serial_sampler import SingleChainSerialSampler
from .smmala import SMMALA
from eeyore.chains import ChainList
from eeyore.datasets import DataCounter
from eeyore.kernels import MultivariateNormalKernel

class GAMC(SingleChainSerialSampler):
    def __init__(self, model, samplers, theta0=None, dataloader=None, data0=None, counter=None, choose_kernel=None,
        a=10., chain=ChainList()):
        super(GAMC, self).__init__(counter or DataCounter.from_dataloader(dataloader))

        self.dataloader = dataloader

        if choose_kernel is not None:
            self.choose_kernel = choose_kernel
        else:
            self.choose_kernel = lambda n, num_iters : self.sample_exp_decay_indicator(n, num_iters, a)
        self.last_kernel = None
        self.current_kernel = None
        self.offset = 0

        self.keys = ['sample', 'target_val', 'accepted']
        self.chain = chain

        self.sampler_names = [samplers[i][0] for i in range(2)]

        self.init_samplers(model, samplers, theta0, data0 or next(iter(self.dataloader)))

    def init_samplers(self, model, samplers, theta0, data0):
        self.samplers = []
        for i in range(2):
            if samplers[i][0] == 'MetropolisHastings':
                self.samplers.append(MetropolisHastings(
                    copy.deepcopy(model),
                    theta0=theta0,
                    dataloader=None,
                    data0=data0,
                    counter=self.counter,
                    symmetric=True,
                    kernel=MultivariateNormalKernel(
                        torch.zeros(model.num_params(), dtype=model.dtype, device=model.device),
                        scale_tril=torch.eye(model.num_params(), dtype=model.dtype, device=model.device)
                    ),
                    chain=self.chain
                ))
            if samplers[i][0] == 'AM':
                self.samplers.append(AM(
                    copy.deepcopy(model),
                    theta0=theta0, dataloader=None, data0=data0, counter=self.counter,
                    chain=self.chain, **(samplers[i][1])
                ))
            if samplers[i][0] == 'RAM':
                self.samplers.append(RAM(
                    copy.deepcopy(model),
                    theta0=theta0, dataloader=None, data0=data0, counter=self.counter,
                    chain=self.chain, **(samplers[i][1])
                ))
            elif samplers[i][0] == 'MALA':
                self.samplers.append(MALA(
                    copy.deepcopy(model),
                    theta0=theta0, dataloader=None, data0=data0, counter=self.counter,
                    chain=self.chain, **(samplers[i][1])
                ))
            elif samplers[i][0] == 'SMMALA':
                self.samplers.append(SMMALA(
                    copy.deepcopy(model),
                    theta0=theta0, dataloader=None, data0=data0, counter=self.counter,
                    chain=self.chain, **(samplers[i][1])
                ))

    def default_indicator(self):
        return self.current_kernel

    def get_model(self, idx=None):
        return self.samplers[idx or self.default_indicator()].model

    def current_sampler(self):
        return self.samplers[self.current_kernel]

    def exp_decay_indicator_prob(self, n, num_iters, a=10.):
        return torch.exp(torch.tensor([- a * n / num_iters]))

    def sample_exp_decay_indicator(self, n, num_iters, a=10.):
        return Bernoulli(self.exp_decay_indicator_prob(n, num_iters, a=a)).sample().to(dtype=torch.int)

    def set_kernel_indicators(self, n, num_iters):
        if self.current_kernel is not None:
            self.last_kernel = self.current_kernel

        self.current_kernel = self.choose_kernel(n, num_iters)

        if self.current_kernel == 1:
            self.offset = n - 1

    def set_all(self, theta, data=None):
        x, y = data or next(iter(self.dataloader))
        for sampler in self.samplers:
            sampler.set_all(theta, data=(x, y))

    def reset_in_sampler(self, theta, data=None, sampler_id=None, reset_counter=True, reset_chain=True):
        self.samplers[sampler_id or self.current_kernel].reset(
            theta, data=data or next(iter(self.dataloader)), reset_counter=reset_counter, reset_chain=reset_chain
        )

    def reset_in_sampler_from_data(self, x, y, reset_counter=True, reset_chain=True):
        self.set_kernel_indicators(self.counter.idx+1, self.counter.num_iters)

        if (self.counter.idx > 0) and (self.current_kernel != self.last_kernel):
            self.reset_in_sampler(
                self.samplers[self.last_kernel].current['sample'].clone().detach(),
                data=(x, y),
                reset_counter=reset_counter,
                reset_chain=reset_chain
            )

    def reset(self, theta, data=None, reset_counter=True, reset_chain=True):
        super().reset(theta, data=data, reset_counter=reset_counter, reset_chain=reset_chain)
        self.last_kernel = None
        self.current_kernel = None
        self.offset = 0

    def draw(self, x, y, savestate=False):
        self.reset_in_sampler_from_data(x, y, reset_counter=False, reset_chain=False)

        if ((self.sampler_names[self.current_kernel] == 'AM') or
            (self.sampler_names[self.current_kernel] == 'RAM')
        ):
            self.current_sampler().draw(x, y, savestate=savestate, offset=self.offset)
        else:
            self.current_sampler().draw(x, y, savestate=savestate)
